/*----------------------------------------------------------------------------*/
/* Copyright (c) FIRST 2008. All Rights Reserved.                             */
/* Open Source Software - may be modified and shared by FRC teams. The code   */
/* must be accompanied by the FIRST BSD license file in the root directory of */
/* the project.                                                               */
/*----------------------------------------------------------------------------*/

package edu.wpi.first.wpilibj;

/**
 * I2C bus interface class.
 *
 * This class is intended to be used by sensor (and other I2C device) drivers.
 * It probably should not be used directly.
 *
 * It is constructed by calling DigitalModule::GetI2C() on a DigitalModule object.
 */
public class I2C extends SensorBase {

    private DigitalModule m_module;
    private int m_deviceAddress;

    /**
     * Constructor.
     *
     * @param module The Digital Module to which the device is conneted.
     * @param deviceAddress The address of the device on the I2C bus.
     */
    public I2C(DigitalModule module, int deviceAddress) {
        if (module == null)
            throw new NullPointerException("Digital Module given was null");
        m_module = module;
        m_deviceAddress = deviceAddress;
    }

    /**
     * Destructor.
     */
    protected void free() {
    }

    /**
     * Execute a write transaction with the device.
     *
     * Write a byte to a register on a device and wait until the
     *   transaction is complete.
     *
     * @param registerAddress The address of the register on the device to be written.
     * @param data The byte to write to the register on the device.
     */
    public synchronized void write(int registerAddress, int data) {
        m_module.m_fpgaDIO.writeI2CConfig_Address(m_deviceAddress);
        m_module.m_fpgaDIO.writeI2CConfig_BytesToWrite(1 + 1); //sizeof registerAdrress + data
        m_module.m_fpgaDIO.writeI2CConfig_BytesToRead(0);
        m_module.m_fpgaDIO.writeI2CDataToSend((data << 8) | registerAddress);
        byte transaction = (byte) m_module.m_fpgaDIO.readI2CStatus_Transaction();
        m_module.m_fpgaDIO.strobeI2CStart();
        while (transaction == m_module.m_fpgaDIO.readI2CStatus_Transaction()) {
            Timer.delay(.001);
        }
        while (!m_module.m_fpgaDIO.readI2CStatus_Done()) {
            Timer.delay(.001);
        }
    }

    /**
     * Execute a read transaction with the device.
     *
     * Read 1, 2, 3, or 4 bytes from a device.
     * Most I2C devices will auto-increment the register pointer internally
     *   allowing you to read up to 4 consecutive registers on a device in a
     *   single transaction.
     *
     * @todo possibly change to allocate the byte array
     *
     * @param registerAddress The register to read first in the transaction.
     * @param count The number of bytes to read in the transaction.
     * @param buffer A pointer to the array of bytes to store the data read from the device.
     */
    public synchronized void read(int registerAddress, int count, byte[] buffer) {
        long data = 0;
        if (count < 1 || count > 4) {
            throw new IndexOutOfBoundsException("The number of bytes to read must be between 1 and 4, inclusive");
        }
        if (buffer == null) {
            throw new NullPointerException("Null byte buffer provided");
        }

        m_module.m_fpgaDIO.writeI2CConfig_Address(m_deviceAddress);
        m_module.m_fpgaDIO.writeI2CConfig_BytesToWrite(1); //size of registerAddress
        m_module.m_fpgaDIO.writeI2CConfig_BytesToRead(count);
        m_module.m_fpgaDIO.writeI2CDataToSend(registerAddress);
        byte transaction = (byte) m_module.m_fpgaDIO.readI2CStatus_Transaction();
        m_module.m_fpgaDIO.strobeI2CStart();
        while (transaction == m_module.m_fpgaDIO.readI2CStatus_Transaction()) {
            Timer.delay(.001);
        }
        while (!m_module.m_fpgaDIO.readI2CStatus_Done()) {
            Timer.delay(.001);
        }
        data = m_module.m_fpgaDIO.readI2CDataReceived();

        for (int i = 0; i < count; i++) {
            buffer[i] = (byte) ((data >> (8 * i)) & 0xFF);
        }
    }

    /**
     * Send a broadcast write to all devices on the I2C bus.
     *
     * This is not currently implemented!
     *
     * @param registerAddress The register to write on all devices on the bus.
     * @param data The value to write to the devices.
     */
    public void broadcast(int registerAddress, int data) {
    }

    /**
     * Verify that a device's registers contain expected values.
     *
     * Most devices will have a set of registers that contain a known value that
     *   can be used to identify them.  This allows an I2C device driver to easily
     *   verify that the device contains the expected value.
     *
     * @pre The device must support and be configured to use register auto-increment.
     *
     * @param registerAddress The base register to start reading from the device.
     * @param count The size of the field to be verified.
     * @param expected A buffer containing the values expected from the device.
     * @return true if the sensor was verified to be connected
     */
    public boolean verifySensor(int registerAddress, int count, byte[] expected) {
        byte[] deviceData = new byte[4];
        for (byte i = 0, curRegisterAddress = (byte) registerAddress; i < count; i += 4, curRegisterAddress += 4) {
            byte toRead = (byte) (count - i < 4 ? count - i : 4);
            read(curRegisterAddress, toRead, deviceData);

            for (byte j = 0; j < toRead; j++) {
                if (deviceData[j] != expected[i + j]) {
                    return false;
                }
            }
        }
        return true;
    }
}
