/*----------------------------------------------------------------------------*/
/* Copyright (c) FIRST 2008. All Rights Reserved.                             */
/* Open Source Software - may be modified and shared by FRC teams. The code   */
/* must be accompanied by the FIRST BSD license file in the root directory of */
/* the project.                                                               */
/*----------------------------------------------------------------------------*/

package edu.wpi.first.wpilibj;

import edu.wpi.first.wpilibj.fpga.tDIO;

/**
 * Class to read a digital input.
 * This class will read digital inputs and return the current value on the channel. Other devices
 * such as encoders, gear tooth sensors, etc. that are implemented elsewhere will automatically
 * allocate digital inputs and outputs as required. This class is only for devices like switches
 * etc. that aren't implemented anywhere else.
 */
public class DigitalOutput extends SensorBase {

    private int m_channel;
    private DigitalModule m_module;

    private void initDigitalOutput(int slot, int channel) {
        m_channel = channel;
        m_module = DigitalModule.getInstance(slot);
        m_module.allocateDIO(m_channel, false);
    }

    /**
     * Create an instance of a digital output.
     * Create an instance of a digital output given a slot and channel.
     * @param slot the slot the digital module is in
     * @param channel the port to use for the digital output
     */
    public DigitalOutput(int slot, int channel) {
        initDigitalOutput(slot, channel);
    }

    /**
     * Create an instance of a digital output.
     * Create a digital output given a channel. The default module is used.
     * @param channel the port to use for the digital output
     */
    public DigitalOutput(int channel) {
        initDigitalOutput(getDefaultDigitalModule(), channel);
    }

    /**
     * Free the resources associated with a digital output.
     */
    protected void free() {
        m_module.freeDIO(m_channel);
    }

    /**
     * Set the value of a digital output.
     * @param value true is on, off is false
     */
    public void set(boolean value) {
        m_module.setDIO(m_channel, value);
    }

    /**
     * Output a single pulse on the digital output line.
     * Send a single pulse on the digital output line where the pulse diration is specified in seconds.
     * Maximum pulse length is 0.0016 seconds.
     * @param length The pulselength in seconds
     */
    public void pulse(double length) {
        m_module.pulse(m_channel, (char) (1e9 * length / (tDIO.readLoopTiming() * 25)));
    }

    /**
     * Determine if the pulse is still going.
     * Determine if a previously started pulse is still going.
     * @return true if pulsing
     */
    public boolean isPulsing() {
        return m_module.isPulsing(m_channel);
    }
}
