/*----------------------------------------------------------------------------*/
/* Copyright (c) FIRST 2008. All Rights Reserved.                             */
/* Open Source Software - may be modified and shared by FRC teams. The code   */
/* must be accompanied by the FIRST BSD license file in the root directory of */
/* the project.                                                               */
/*----------------------------------------------------------------------------*/
package edu.wpi.first.wpilibj;

import edu.wpi.first.wpilibj.communication.*;

/**
 * Provide access to the network communication data to / from the Driver Station.
 */
public class DriverStation {

    private class DriverStationTask implements Runnable {

        private DriverStation m_ds;

        public DriverStationTask(DriverStation ds) {
            m_ds = ds;
        }

        public void run() {
            m_ds.task();
        }
    }
    /**
     * The size of the user control data
     */
    public static final int USER_CONTROL_DATA_SIZE = FRCControl.USER_CONTROL_DATA_SIZE;
    /**
     * The size of the user status data
     */
    public static final int USER_STATUS_DATA_SIZE = FRCControl.USER_STATUS_DATA_SIZE;
    private static final int WAIT_FOREVER = -1;
    /**
     * Slot for the analog module to read the battery
     */
    public static final int kBatterySlot = 1;
    /**
     * Analog channel to read the battery
     */
    public static final int kBatteryChannel = 8;
    /**
     * Number of Joystick Ports
     */
    public static final int kJoystickPorts = 4;
    /**
     * Number of Joystick Axes
     */
    public static final int kJoystickAxes = 6;

    /**
     * The robot alliance that the robot is a part of
     */
    public static class Alliance {

        /**
         * The integer value representing this enumeration
         */
        public final int value;
        static final int kRed_val = 0;
        static final int kBlue_val = 1;
        static final int kInvalid_val = 2;
        /**
         * alliance: Red
         */
        public static final Alliance kRed = new Alliance(kRed_val);
        /**
         * alliance: Blue
         */
        public static final Alliance kBlue = new Alliance(kBlue_val);
        /**
         * alliance: Invalid
         */
        public static final Alliance kInvalid = new Alliance(kInvalid_val);

        private Alliance(int value) {
            this.value = value;
        }
    }
    private static final double kUpdatePeriod = .050;
    private static DriverStation instance = new DriverStation();
    private FRCCommonControlData m_controlData;
    private AnalogChannel m_batteryChannel;
    private Thread m_thread;
    private final Object m_semaphore;
    private int m_digitalOut;
    private boolean m_thread_keepalive = true;
    private Dashboard m_dashboardHigh;
    private Dashboard m_dashboardLow;

    private DriverStationEnhancedIO m_enhancedIO = new DriverStationEnhancedIO();

    /**
     * Gets an instance of the DriverStation
     *
     * @return The DriverStation.
     */
    public static DriverStation getInstance() {
        return DriverStation.instance;
    }

    /**
     * DriverStation contructor.
     *
     * The single DriverStation instance is created statically with the
     * instance static member variable.
     */
    protected DriverStation() {
        m_controlData = new FRCCommonControlData();
        m_semaphore = new Object();

        m_dashboardHigh = new Dashboard(m_semaphore);
        m_dashboardLow = new Dashboard(m_semaphore);

        // initialize packet number and control words to zero;
        m_controlData.packetIndex = 0;
        m_controlData.control = 0;

        // set all joystick axis values to neutral; buttons to OFF
        m_controlData.stick0Axes = new byte[]{0, 0, 0, 0, 0, 0};
        m_controlData.stick1Axes = new byte[]{0, 0, 0, 0, 0, 0};
        m_controlData.stick2Axes = new byte[]{0, 0, 0, 0, 0, 0};
        m_controlData.stick3Axes = new byte[]{0, 0, 0, 0, 0, 0};
        m_controlData.stick0Buttons = 0;
        m_controlData.stick1Buttons = 0;
        m_controlData.stick2Buttons = 0;
        m_controlData.stick3Buttons = 0;

        // initialize the analog and digital data.
        m_controlData.analog1 = 0;
        m_controlData.analog2 = 0;
        m_controlData.analog3 = 0;
        m_controlData.analog4 = 0;
        m_controlData.dsDigitalIn = 0;

        m_batteryChannel = new AnalogChannel(kBatterySlot, kBatteryChannel);

        m_thread = new Thread(new DriverStationTask(this), "FRCDriverStation");

        m_thread.start();


    }

    /**
     * Kill the thread
     */
    public void release() {
        m_thread_keepalive = false;
    }

    /**
     * Provides the service routine for the DS polling thread.
     */
    private void task() {
        while (m_thread_keepalive) {
            getData();
            m_enhancedIO.updateData();
            setData();
            Timer.delay(kUpdatePeriod);
        }
    }

    /**
     * Copy data from the DS task for the user.
     * If no new data exists, it will just be returned, otherwise
     * the data will be copied from the DS polling loop.
     */
    protected void getData() {
        FRCControl.getCommonControlData(m_controlData, WAIT_FOREVER);
    }

    /**
     * Copy status data from the DS task for the user.
     * This is used primarily to set digital outputs on the DS.
     */
    protected void setData() {
        synchronized (m_semaphore) {
            FRCControl.setStatusData(getBatteryVoltage(), m_digitalOut,
                    Dashboard.s_updateNumber, m_dashboardHigh.m_userStatus.m_bytes,
                    m_dashboardHigh.m_userStatusSize,
                    m_dashboardLow.m_userStatus.m_bytes,
                    m_dashboardLow.m_userStatusSize,
                    WAIT_FOREVER);
        }
    }

    /**
     * Read the battery voltage from the specified AnalogChannel.
     *
     * This accessor assumes that the battery voltage is being measured
     * through the voltage divider on an analog breakout.
     *
     * @return The battery voltage.
     */
    public double getBatteryVoltage() {
        // The Analog bumper has a voltage divider on the battery source.
        // Vbatt *--/\/\/\--* Vsample *--/\/\/\--* Gnd
        //         680 Ohms            1000 Ohms
        return m_batteryChannel.getAverageVoltage() * (1680.0 / 1000.0);
    }

    /**
     * Get the value of the axis on a joystick.
     * This depends on the mapping of the joystick connected to the specified port.
     *
     * @param stick The joystick to read.
     * @param axis The analog axis value to read from the joystick.
     * @return The value of the axis on the joystick.
     */
    public double getStickAxis(int stick, int axis) {
        if (axis < 1 || axis > kJoystickAxes) {
            return 0.0;
        }

        int value;
        getData();
        switch (stick) {
            case 1:
                value = m_controlData.stick0Axes[axis - 1];
                break;
            case 2:
                value = m_controlData.stick1Axes[axis - 1];
                break;
            case 3:
                value = m_controlData.stick2Axes[axis - 1];
                break;
            case 4:
                value = m_controlData.stick3Axes[axis - 1];
                break;
            default:
                return 0.0;
        }

        return (value) / 128.0;
    }

    /**
     * The state of the buttons on the joystick.
     * 12 buttons (4 msb are unused) from the joystick.
     *
     * @param stick The joystick to read.
     * @return The state of the buttons on the joystick.
     */
    public int getStickButtons(final int stick) {
        getData();
        switch (stick) {
            case 1:
                return m_controlData.stick0Buttons;
            case 2:
                return m_controlData.stick1Buttons;
            case 3:
                return m_controlData.stick2Buttons;
            case 4:
                return m_controlData.stick3Buttons;
            default:
                return 0;
        }
    }

    /**
     * Get an analog voltage from the Driver Station.
     * The analog values are returned as int values for the Driver Station analog inputs.
     * These inputs are typically used for advanced operator interfaces consisting of potentiometers
     * or resistor networks representing values on a rotary switch.
     *
     * @param channel The analog input channel on the driver station to read from. Valid range is 1 - 4.
     * @return The analog voltage on the input.
     */
    public double getAnalogIn(final int channel) {
        getData();
        switch (channel) {
            case 1:
                return m_controlData.analog1;
            case 2:
                return m_controlData.analog2;
            case 3:
                return m_controlData.analog3;
            case 4:
                return m_controlData.analog4;
            default:
                return 0.0;
        }
    }

    /**
     * Get values from the digital inputs on the Driver Station.
     * Return digital values from the Drivers Station. These values are typically used for buttons
     * and switches on advanced operator interfaces.
     * @param channel The digital input to get. Valid range is 1 - 8.
     * @return The value of the digital input
     */
    public boolean getDigitalIn(final int channel) {
        getData();
        return ((m_controlData.dsDigitalIn >> (channel - 1)) & 0x1) == 0x1;
    }

    /**
     * Set a value for the digital outputs on the Driver Station.
     *
     * Control digital outputs on the Drivers Station. These values are typically used for
     * giving feedback on a custom operator station such as LEDs.
     *
     * @param channel The digital output to set. Valid range is 1 - 8.
     * @param value The state to set the digital output.
     */
    public void setDigitalOut(final int channel, final boolean value) {
        m_digitalOut &= ~(0x1 << (channel - 1));
        m_digitalOut |= ((value ? 1 : 0) << (channel - 1));
    }

    /**
     * Get a value that was set for the digital outputs on the Driver Station.
     * @param channel The digital ouput to monitor. Valid range is 1 through 8.
     * @return A digital value being output on the Drivers Station.
     */
    public boolean getDigitalOut(final int channel) {
        return ((m_digitalOut >> (channel - 1)) & 0x1) == 0x1;
    }

    /**
     * Gets a value indicating whether the Driver Station requires the
     * robot to be enabled.
     *
     * @return True if the robot is enabled, false otherwise.
     */
    public boolean isEnabled() {
        getData();
        return m_controlData.enabled();
    }

    /**
     * Gets a value indicating whether the Driver Station requires the
     * robot to be disabled.
     *
     * @return True if the robot should be disabled, false otherwise.
     */
    public boolean isDisabled() {
        getData();
        return !m_controlData.enabled();
    }

    /**
     * Gets a value indicating whether the Driver Station requires the
     * robot to be running in autonomous mode.
     *
     * @return True if autonomous mode should be enabled, false otherwise.
     */
    public boolean isAutonomous() {
        getData();
        return m_controlData.autonomous();
    }

    /**
     * Gets a value indicating whether the Driver Station requires the
     * robot to be running in operator-controlled mode.
     *
     * @return True if operator-controlled mode should be enabled, false otherwise.
     */
    public boolean isOperatorControl() {
        getData();
        return !m_controlData.autonomous();
    }

    /**
     * Return the DS packet number.
     * The packet number is the index of this set of data returned by the driver station.
     * Each time new data is received, the packet number (included with the sent data) is returned.
     *
     * @return The DS packet number.
     */
    public int getPacketNumber() {
        getData();
        return m_controlData.packetIndex;
    }

    /**
     * Get the current alliance from the FMS
     * @return the current alliance
     */
    public Alliance getAlliance() {
        switch (m_controlData.dsID_Alliance) {
            case 'R':
                return Alliance.kRed;
            case 'B':
                return Alliance.kBlue;
            default:
                return Alliance.kInvalid;
        }
    }

    /**
     * Get the location of the controls
     * @return the location of the controls
     */
    public int getLocation() {
        return m_controlData.dsID_Position - '0';
    }

    /**
     * Get the dashboard packer for sending high user data to a dashboard program
     * @return a Dashboard object
     */
    public Dashboard getDashboardPackerHigh() {
        return m_dashboardHigh;
    }

    /**
     * Get the dashboard packer for sending low priority user data to a dashboard program
     * @return a Dashboard object
     */
    public Dashboard getDashboardPackerLow() {
        return m_dashboardLow;
    }

    /**
     * Is the driver station attached to a Field Management System?
     * Note: This does not work with the Blue DS.
     * @return True if the robot is competing on a field being controlled by a Field Management System
     */
    boolean isFMSAttached() {
        return (m_controlData.control & FRCCommonControlData.FMS_ATTATCHED) > 0;
    }

    /**
     * Get the interface to the enhanced IO of the new driver station.
     * @return An enhanced IO object for the advanced features of the driver
     * station.
     */
    public DriverStationEnhancedIO getEnhancedIO () {
        return m_enhancedIO;
    }
}
