/*----------------------------------------------------------------------------*/
/* Copyright (c) FIRST 2008. All Rights Reserved.                             */
/* Open Source Software - may be modified and shared by FRC teams. The code   */
/* must be accompanied by the FIRST BSD license file in the root directory of */
/* the project.                                                               */
/*----------------------------------------------------------------------------*/

package edu.wpi.first.wpilibj;

import edu.wpi.first.wpilibj.fpga.tSolenoid;
import edu.wpi.first.wpilibj.util.AllocationException;
import edu.wpi.first.wpilibj.util.CheckedAllocationException;

/**
 * Solenoid class for running high voltage Digital Output (9472 module).
 *
 * The Solenoid class is typically used for pneumatics solenoids, but could be used
 * for any device within the current spec of the 9472 module.
 */
public class Solenoid extends SensorBase {

    private static Resource allocated = new Resource(tSolenoid.kDO7_0_NumElements * kSolenoidChannels);
    private int m_chassisSlot; ///< Slot number where the module is plugged into the chassis.
    private int m_channel; ///< The channel on the module to control.
    private static tSolenoid m_fpgaSolenoidModule; ///< FPGA Solenoid Module object.
    private static int m_refCount; ///< Reference count for the chip object.

    /**
     * Common function to implement constructor behavior.
     */
    private synchronized void initSolenoid() {
        checkSolenoidModule(m_chassisSlot);
        checkSolenoidChannel(m_channel);

        m_refCount++;
        if (m_refCount == 1) {
            m_fpgaSolenoidModule = new tSolenoid();
        }

        try {
            allocated.allocate(slotToIndex(m_chassisSlot) * kSolenoidChannels + m_channel - 1);
        } catch (CheckedAllocationException e) {
            throw new AllocationException(
                    "Solenoid channel " + m_channel + " on module " + m_chassisSlot + " is already allocated");
        }
    }

    /**
     * Constructor.
     *
     * @param channel The channel on the module to control.
     */
    public Solenoid(final int channel) {
        m_chassisSlot = getDefaultSolenoidModule();
        m_channel = channel;
        initSolenoid();
    }

    /**
     * Constructor.
     *
     * @param slot The slot that the 9472 module is plugged into.
     * @param channel The channel on the module to control.
     */
    public Solenoid(final int slot, final int channel) {
        m_chassisSlot = slot;
        m_channel = channel;
        initSolenoid();
    }

    /**
     * Destructor.
     */
    protected synchronized void free() {
        allocated.free(slotToIndex(m_chassisSlot) * kSolenoidChannels + m_channel - 1);
        if (m_refCount == 1) {
            m_fpgaSolenoidModule.Release();
            m_fpgaSolenoidModule = null;
        }
        m_refCount--;
    }

    /**
     * Convert slot number to index.
     *
     * @param slot The slot in the chassis where the module is plugged in.
     * @return An index to represent the module internally.
     */
    static public int slotToIndex(final int slot) {
        return 8 - slot;
    }

    /**
     * Set the value of a solenoid.
     *
     * @param on Turn the solenoid output off or on.
     */
    public synchronized void set(boolean on) {
        {
            int value = tSolenoid.readDO7_0(slotToIndex(m_chassisSlot));
            if (on) {
                value |= 1 << (m_channel - 1);
            } else {
                value &= ~(1 << (m_channel - 1));
            }
            tSolenoid.writeDO7_0(slotToIndex(m_chassisSlot), value);
        }

    }

    /**
     * Read the current value of the solenoid.
     *
     * @return The current value of the solenoid.
     */
    public boolean get() {
        int value = tSolenoid.readDO7_0(slotToIndex(m_chassisSlot)) & (1 << (m_channel - 1));
        return (value != 0);
    }

    /**
     * Read the current value of all solenoids.
     *
     * @return The current value of all solenoids.
     */
    public static byte getAll() {
        return (byte)tSolenoid.readDO7_0(slotToIndex(getDefaultSolenoidModule()));
    }
}
